from typing import Optional
from functools import lru_cache
from emoji.unicode_codes.data_dict import EMOJI_DATA, STATUS, LANGUAGES

__all__ = [
    'get_emoji_by_name',
    'EMOJI_DATA', 'STATUS', 'LANGUAGES'
]


@lru_cache(maxsize=4000)
def get_emoji_by_name(name: str, language: str) -> Optional[str]:
    """
    Find emoji by short-name in a specific language.
    Returns None if not found

    :param name: emoji short code e.g. ":banana:"
    :param language: language-code e.g. 'es', 'de', etc. or 'alias'
    """

    fully_qualified = STATUS['fully_qualified']

    if language == 'alias':
        for emj, data in EMOJI_DATA.items():
            if name in data.get('alias', []) and data['status'] <= fully_qualified:
                return emj
        language = 'en'

    for emj, data in EMOJI_DATA.items():
        if data.get(language) == name and data['status'] <= fully_qualified:
            return emj

    return None



from typing import List
import random
import pytest
import functools


def pytest_sessionstart():
    # Increase cache size to unlimited size to avoid cache misses during tests
    import emoji.unicode_codes
    emoji.unicode_codes.get_emoji_by_name = functools.lru_cache(
        maxsize=None)(emoji.unicode_codes.get_emoji_by_name.__wrapped__)


def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--shuffle", dest="shuffle", action='store_true',
                     default=False, help="Run tests in random order")


def pytest_collection_modifyitems(session: pytest.Session, items: List[pytest.Item]):
    if session.config.getoption("shuffle"):
        print("")
        print("Shuffling items for a random test order")
        random.shuffle(items)



"""Unittests for versions in EMOJI_DATA"""

from typing import Any, Dict, List
import pytest
import emoji.unicode_codes
from testutils import all_language_packs

def test_emoji_versions_complete_emojize():
    # Check that every emoji has a valid version
    replacement = "<3"
    for lang_code, emoji_pack in all_language_packs():
        for name in emoji_pack.keys():
            version: List[float] = []

            def f(e: str, d: Dict[str, Any]) -> str:
                v = d['E']
                n = d[lang_code]
                assert n == name
                assert isinstance(v, (int, float))
                assert v >= 0.6
                version.append(v)
                return replacement
            r = emoji.emojize(name, language=lang_code,
                              version=0.0, handle_version=f)
            assert len(version) == 1
            assert r == replacement


def test_emoji_versions_complete_demojize():
    # Check that every emoji has a valid version
    for lang_code, emoji_pack in all_language_packs():
        for name in emoji_pack.keys():
            version: List[float] = []

            def f(e: str, d: Dict[str, Any]) -> str:
                v = d['E']
                assert isinstance(v, (int, float))
                assert v >= 0.6
                version.append(v)
                return ''
            emoji.demojize(emoji.emojize(name, language=lang_code),
                           language=lang_code, version=0.0, handle_version=f)
            if len(version) != 1:
                print(name)
                print(emoji.emojize(name, language=lang_code).encode("unicode-escape").decode("utf-8"))
                print(emoji.demojize(emoji.emojize(name, language=lang_code), language=lang_code).encode("unicode-escape").decode("utf-8"))
            assert len(version) == 1


def test_method_version():
    # Test method "emoji.version()"

    assert emoji.version(":snake:") == 0.6
    assert emoji.version("\U0001F40D") == 0.6

    assert emoji.version(":brain:") == 5
    assert emoji.version("\U0001F9E0") == 5
    assert emoji.version(":cerebro:") == 5

    assert emoji.version("prefix :people_hugging: suffix") == 13
    assert emoji.version("prefix \U0001FAC2 suffix") == 13

    assert emoji.version("\U0001F611") == 1
    assert emoji.version(":expressionless_face:") == 1
    assert emoji.version(":expressionless:") == 1

    assert emoji.version("':pouring_liquid::people_hugging:") == 14
    assert emoji.version("\U0001FAD7\U0001FAC2") == 14

    assert emoji.emojize(':123:', version=5, variant="text_type") == ':123:'

    with pytest.raises(ValueError):
        emoji.version("test")

    with pytest.raises(ValueError):
        emoji.version("':snak:")


def test_method_replace_version():
    # Test method "emoji.replace_emoji(string, version=X.Y)"

    assert emoji.replace_emoji("\U0001F40D\U0001F9E0\U0001FAC2\U0001FAD7\U0001FAC2", version=6) == "\U0001F40D\U0001F9E0"

    assert emoji.replace_emoji('Hi, I am fine. 😁', version=0) == 'Hi, I am fine. '
    assert emoji.replace_emoji('Hi', version=0) == 'Hi'
    assert emoji.replace_emoji('Hello 🇫🇷👌', version=0) == 'Hello '
    assert emoji.replace_emoji('Hello 🇫🇷👌', 'x', version=0,) == 'Hello xx'
    assert emoji.replace_emoji('Hello 🇫🇷👌', 'x', version=1,) == 'Hello 🇫🇷👌'

    def replace(emj: str, data: Dict[str, Any]) -> str:
        assert emj in ["🇫🇷", "👌"]
        return 'x'
    assert emoji.replace_emoji('Hello 🇫🇷👌', replace, version=0.1) == 'Hello xx'

    assert emoji.replace_emoji('A 🦖 is eating a 🥐', replace='', version=5.0) == 'A 🦖 is eating a 🥐'



"""Tests for emoji that consist of multiple emoji joined with a u200D (ZWJ - zero width joiner)
This file contains tests that are irrespective of keeping/removing the ZWJ.
See test_zwj_remove.py for tests when the ZWJ is removed.
See test_zwj_keep.py for tests when the ZWJ is kept.
"""

import emoji


def test_non_rgi_zwj_emoji_list():
    matches = emoji.emoji_list('\U0001F468\u200d\U0001F469\U0001F3FF\u200d\U0001F467\U0001F3FB\u200d\U0001F466\U0001F3FE')

    assert matches[0]["match_start"] == 0
    assert matches[0]["match_end"] == 1
    assert matches[0]["emoji"] == '\U0001F468'

    assert matches[1]["match_start"] == 2
    assert matches[1]["match_end"] == 4
    assert matches[1]["emoji"] == '\U0001F469\U0001F3FF'

    assert matches[2]["match_start"] == 5
    assert matches[2]["match_end"] == 7
    assert matches[2]["emoji"] == '\U0001F467\U0001F3FB'

    assert matches[3]["match_start"] == 8
    assert matches[3]["match_end"] == 10
    assert matches[3]["emoji"] == '\U0001F466\U0001F3FE'

    matches = emoji.emoji_list('abc\U0001F468\u200d\U0001F469\U0001F3FF\u200d\U0001F467\U0001F3FB\u200d\U0001F466\U0001F3FEyxz')

    assert matches[0]["match_start"] == 3
    assert matches[0]["match_end"] == 4
    assert matches[0]["emoji"] == '\U0001F468'

    assert matches[1]["match_start"] == 5
    assert matches[1]["match_end"] == 7
    assert matches[1]["emoji"] == '\U0001F469\U0001F3FF'

    assert matches[2]["match_start"] == 8
    assert matches[2]["match_end"] == 10
    assert matches[2]["emoji"] == '\U0001F467\U0001F3FB'

    assert matches[3]["match_start"] == 11
    assert matches[3]["match_end"] == 13
    assert matches[3]["emoji"] == '\U0001F466\U0001F3FE'

    matches = emoji.emoji_list('\U0001F9D1\U0000200D\U0001F9B3abcyxz\U0001FAC5\U0001F468\U0001F3FD\u200D\U0001F467\U0001F3FC\u200D\U0001F467\U0001F3FE')

    assert matches[0]["match_start"] == 0
    assert matches[0]["match_end"] == 3
    assert matches[0]["emoji"] == '\U0001F9D1\U0000200D\U0001F9B3'

    assert matches[1]["match_start"] == 9
    assert matches[1]["match_end"] == 10
    assert matches[1]["emoji"] == '\U0001FAC5'

    assert matches[2]["match_start"] == 10
    assert matches[2]["match_end"] == 12
    assert matches[2]["emoji"] == '\U0001F468\U0001F3FD'

    assert matches[3]["match_start"] == 13
    assert matches[3]["match_end"] == 15
    assert matches[3]["emoji"] == '\U0001F467\U0001F3FC'

    assert matches[4]["match_start"] == 16
    assert matches[4]["match_end"] == 18
    assert matches[4]["emoji"] == '\U0001F467\U0001F3FE'


def test_non_rgi_zwj_demojize():
    result = emoji.demojize('\U0001F9D1\U0001F3FC\U0000200D\U0001F3A8\U0001F468\u200d\U0001F469\U0001F3FF\u200d\U0001F467\U0001F3FB\u200d\U0001F466\U0001F3FE')
    assert '\U0001F9D1\U0001F3FC\U0000200D\U0001F3A8' not in result
    assert '\U0001F468\u200d\U0001F469\U0001F3FF\u200d\U0001F467\U0001F3FB\u200d\U0001F466\U0001F3FE' not in result
    assert ':artist_medium-light_skin_tone:' in result

    result = emoji.demojize('Test \U0001F6B5\U0001F3FF\U0000200D\U00002642\U0000FE0F abc \U0001F468\U0001F3FF\u200d\U0001F469\U0001F3FE\u200d\U0001F466\U0001F3FD\u200d\U0001F467\U0001F3FB')
    assert '\U0001F6B5\U0001F3FF\U0000200D\U00002642\U0000FE0F' not in result
    assert '\U0001F468\U0001F3FF\u200d\U0001F469\U0001F3FE\u200d\U0001F466\U0001F3FD\u200d\U0001F467\U0001F3FB' not in result
    assert ':man_mountain_biking_dark_skin_tone:' in result


def test_malformed_zwj_no_emoji():
    s = "\u200D"
    assert emoji.replace_emoji(s) == s

    s = "\u200D\u200D"
    assert emoji.replace_emoji(s) == s

    s = "\u200D\u200D\u200D"
    assert emoji.replace_emoji(s) == s

    s = "Has\u200Din the middle"
    assert emoji.replace_emoji(s) == s

    s = "\u200DStarts With"
    assert emoji.replace_emoji(s) == s

    s = "Ends With\u200D"
    assert emoji.replace_emoji(s) == s

    s = "Multiple\u200D\u200D\u200Din the middle"
    assert emoji.replace_emoji(s) == s

    s = "\u200D\u200DStarts With two"
    assert emoji.replace_emoji(s) == s

    s = "\u200D\u200D\u200DStarts With three"
    assert emoji.replace_emoji(s) == s

    s = "Ends With two\u200D\u200D"
    assert emoji.replace_emoji(s) == s

    s = "Ends With three\u200D\u200D\u200D"
    assert emoji.replace_emoji(s) == s


def test_malformed_zwj_mixed_with_emoji():
    emoji.config.demojize_keep_zwj = True  # Restore default config value

    i = "Has🦷\u200Din the middle"
    o = "Has:tooth:\u200Din the middle"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Has\u200D🦷in the middle"
    o = "Has\u200D:tooth:in the middle"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D🦷Starts with"
    o = "\u200D:tooth:Starts with"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "🦷\u200DStarts with"
    o = ":tooth:\u200DStarts with"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Ends with \u200D🦷"
    o = "Ends with \u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Ends with 🦷\u200D"
    o = "Ends with :tooth:\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Multiple 🦷\u200D🦷\u200D in the middle"
    o = "Multiple :tooth:\u200D:tooth:\u200D in the middle"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Multiple 🦷🦷\u200D\u200D in the middle"
    o = "Multiple :tooth::tooth:\u200D\u200D in the middle"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Multiple \u200D\u200D🦷🦷 in the middle"
    o = "Multiple \u200D\u200D:tooth::tooth: in the middle"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D🦷Starts with two"
    o = "\u200D\u200D:tooth:Starts with two"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D\u200D🦷Starts with three"
    o = "\u200D\u200D\u200D:tooth:Starts with three"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Ends with two \u200D\u200D🦷"
    o =  "Ends with two \u200D\u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Ends with two 🦷\u200D\u200D"
    o =  "Ends with two :tooth:\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"


    i = "Ends with three \u200D\u200D\u200D🦷"
    o =  "Ends with three \u200D\u200D\u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "Ends with three 🦷\u200D\u200D\u200D"
    o =  "Ends with three :tooth:\u200D\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "🦷\u200D"
    o =  ":tooth:\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D🦷"
    o =  "\u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D🦷"
    o =  "\u200D\u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "🦷\u200D\u200D"
    o =  ":tooth:\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D🦷\u200D"
    o =  "\u200D:tooth:\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D🦷\u200D\u200D"
    o =  "\u200D\u200D:tooth:\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D\u200D🦷\u200D\u200D"
    o =  "\u200D\u200D\u200D:tooth:\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D🦷\u200D\u200D\u200D"
    o =  "\u200D\u200D:tooth:\u200D\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "🦷\u200D\u200D🦷\u200D\u200D\u200D"
    o =  ":tooth:\u200D\u200D:tooth:\u200D\u200D\u200D"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D🦷🦷\u200D\u200D\u200D🦷"
    o =  "\u200D\u200D:tooth::tooth:\u200D\u200D\u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"

    i = "\u200D\u200D🦷\u200D\u200D\u200D🦷"
    o =  "\u200D\u200D:tooth:\u200D\u200D\u200D:tooth:"
    assert emoji.demojize(i) == o, f"{i!r} != {o!r}"
